
# THIS FILE IS GENERATED FROM THEANO SETUP.PY
short_version = '0.9.0'
version = '0.9.0'
git_revision = 'unknown-git'
full_version = '0.9.0.dev-%(git_revision)s' % {
    'git_revision': git_revision}
release = False

if not release:
    version = full_version
